"""
### CODE OWNERS: Demerrick Moton
### OBJECTIVE:
    Create high-level constructs for implementing and managing ETL processes
### DEVELOPER NOTES:
"""
import logging
import os
from datetime import datetime
import json
from typing import Union
from pyspark.conf import SparkConf
import time
import random

from pyspark import SparkConf
from pyspark.sql import SparkSession, DataFrame
from pyspark.context import SparkContext
from pathlib import Path

from .config import load_config
from .io.db import DBConnector
from .io.bigquery import BigQueryConnector
from .utils.secret_connect import SecretManagerConnector
from .schema import SchemaReader

logging.basicConfig(
    format="%(asctime)s - %(message)s", level=os.environ.get("LOGLEVEL", "INFO")
)
LOGGER = logging.getLogger(__name__)

CURRENT_DIR = Path(__file__).parent
SPARK_BQ_JAR = str(CURRENT_DIR / "io/spark_bq/spark-bigquery-latest_2.12.jar")

# ===========================================================


class SparkApp(SparkSession):
    def __init__(
        self,
        args: str = None,
        conf: SparkConf = None,
        local: bool = False,
        bq_key: str = None,
    ):
        self.sparkapp_params = {
            "master": "yarn",
            "app_name": "",
            "bq_bucket_name": "",
            "project_id": "",
            "secret_id": "",
            "omit_db": False,
            "omit_bq": False,
        }

        if local:
            assert (
                bq_key
            ), "BigQuery API key must be provided in order to use local SparkApp instance"
            super().__init__(self._create_local_spark_context(bq_key=bq_key, conf=conf))
        else:

            if not args:
                LOGGER.warn("No SparkApp arguments were provided")
                return

            # load configuration arguments from 'args'
            self._load_args(args)
            super().__init__(self._create_spark_context(conf))

            # initialize schema reader
            self.schema_reader = SchemaReader()

            # set up BigQuery and Database instance
            self.db_connector = (
                self._setup_db() if not self.sparkapp_params["omit_db"] else None
            )
            self.bq_connector = (
                self._setup_bigquery() if not self.sparkapp_params["omit_bq"] else None
            )

    def __eq__(self, other):
        assert getattr(other, "app_name"), "Object {} is incompatible".format(other)
        assert other.app_name, "Object {} is not properly initialized".format(other)
        return self.app_name == other.app_name

    def __ne__(self, other):
        assert getattr(other, "app_name"), "Object {} is incompatible".format(other)
        assert other.app_name, "Object {} is not properly initialized".format(other)
        return self.app_name != other.app_name

    def _get_rand_num():
        return str(random.randrange(1, 420, 1))

    def _load_args(self, args: str):
        """
        Load dictionary-like string to be parsed into Spark configuration

        Args:
            args (str): dictionary-like string
        """
        args_dict = json.loads(args)
        self.sparkapp_params = {**self.sparkapp_params, **args_dict}

    def _create_spark_context(self, conf: SparkContext):
        LOGGER.info("Creating Spark Context")
        app_name = self.sparkapp_params["app_name"]
        master = self.sparkapp_params["master"]
        rand_num = self._get_rand_num()
        app_name = f"{app_name}_{rand_num}"
        conf = conf or SparkConf()
        conf.set("spark.jars", SPARK_BQ_JAR)
        spark_context = SparkContext(master=master, appName=app_name, conf=conf)
        return spark_context

    def _create_local_spark_context(self, bq_key: str, conf: SparkConf = None):
        """
        Initialize a local SparkApp instance. Useful for quick debugging.

        Args:
            config_file (Union[Path, str]): Path to local configuration file
        Returns:
            ciq_pyspark.SparkApp: SparkApp instance
        """
        LOGGER.info("Creating Spark Context")
        assert Path(
            bq_key
        ).exists(), (
            "BigQuery json key was not found. Cannot start local pyspark instance"
        )

        # check that Spark BQ jar is up to date
        # print current auth and project according to gsutil
        app_name = "local_" + str(random.randrange(1, 420, 1))
        conf = conf or SparkConf()
        conf.set("spark.jars", SPARK_BQ_JAR)
        conf.set("credentialsFile", bq_key)
        spark_context = SparkContext(
            master="local",
            appName=app_name,
            conf=conf,
        )
        return spark_context

    def _setup_bigquery(self):
        LOGGER.info("Initializing BigQuery instance")
        self.conf.set("temporaryGcsBucket", self.sparkapp_params["bq_bucket_name"])
        return BigQueryConnector(self.sparkapp_params["project_id"])

    def _setup_db(self):
        LOGGER.info("Initializing database instance")
        sm_con = SecretManagerConnector()

        # use creds to get values from secret manager
        try:
            creds_vals = sm_con.read_secrets_dictionary(
                self.sparkapp_params["project_id"], self.sparkapp_params["secret_id"]
            )

            return DBConnector(
                instance_ip=creds_vals["db_params"]["private_ip"],
                db_name=creds_vals["db_params"]["dbname"],
                username=creds_vals["db_creds"]["username"],
                password=creds_vals["db_creds"]["password"],
            )
        except Exception as error:
            LOGGER.warn("Unable to initialize Database instance")
            LOGGER.error(error)
            return None

    def load_csv(self, path: str, header: bool = True) -> DataFrame:
        """
        Convert csv file into Pyspark dataframe

        Args:
            path (str): Path to csv file
            header (bool, optional): [description]. Defaults to True.

        Returns:
            DataFrame: Pyspark dataframe
        """
        LOGGER.info("Loading table %s from csv", path)
        _df = self.read.csv(path, header=header)
        LOGGER.info("Table %s loaded from csv", path)
        return _df

    def load_db(
        self, table: str, columns: list = ["*"], limit: int = None
    ) -> DataFrame:
        """
        Convert database table into Pyspark dataframe

        Args:
            table (str): Table name
            columns (list, optional): List is columns to read. Defaults to ["*"].
            limit (int, optional): Maximum number of columns to read at once. Defaults to None.

        Returns:
            DataFrame: [description]
        """
        LOGGER.info("Loading table %s from database", table)
        assert self.db_connector, "Database is not currently connected"
        if limit and limit > 0:
            limit_clause = f" limit {limit}"
            pdf = self.db_connector.read_table(table, columns, limit_clause)
        else:
            pdf = self.db_connector.read_table(table, columns)
        LOGGER.info("Loading schema for table %s", table)
        schema_df = self.db_connector.read_schema_df(table)
        schema = self.schema_reader.generate_schema(schema_df)

        LOGGER.info("Creating dataframe from db table {}".format(table))
        _df = self.createDataFrame(pdf, schema)
        LOGGER.info("Table %s loaded from database", table)
        return _df

    def load_bq(self, dataset: str, table: str) -> DataFrame:
        """
        Convert BigQuery table to a Pyspark dataframe

        Args:
            dataset (str): Name if BigQuery dataset
            table (str): Name of BigQuery table
        Returns:
            DataFrame: Pyspark dataframe
        """
        LOGGER.info("Loading table %s from BigQuery", table)
        _df = (
            self.read.format("bigquery")
            .option("project", f"{self.sparkapp_params['project_id']}")
            .option(
                "table",
                f"{dataset}.{table}",
            )
            .load()
        )
        LOGGER.info("Table %s was loaded from BigQuery", table)
        return _df

    def load_json(self, path: str):
        LOGGER.info("Loading json from {}".format(path))
        _df = self.read.json(path)
        return _df

    def save_csv(self, df: DataFrame, path: str, sep: str = ","):
        LOGGER.info("Writing table {} to csv".format(path))
        df = self.toPandas()
        df.to_csv(path)

    def save_json(self, path):
        LOGGER.info("Writing table {} to json".format(path))
        _df = self.toPandas()
        self.write.json("path")

    def save_fwf(self):
        pass

    def save_excel(self):
        pass

    def save_bq(self, df: DataFrame, dataset: str, table: str, save_mode="overwrite"):
        """
        Save Pyspark dataframe to BigQuery

        Args:
            df (DataFrame): Pyspark dataframe
            dataset (str): Name of BigQuery dataset
            table (str): Name of BigQuery table
            save_mode (str, optional): Spark overwrite policy[append, errorifexists, ignore, overwrite]. Defaults to "overwrite".
        """
        LOGGER.info("Writing table {} to BigQuery".format(table))
        df.write.format("bigquery").option(
            "table", "{}.{}".format(dataset, table)
        ).mode(saveMode=save_mode).save()

    def save_db(self, df: DataFrame, db: DBConnector, table_name: str):
        db.write_table(df, table_name=table_name)

    def view(self, df: DataFrame, nrows: int = 420):
        """
        Launch a PyQT5 table of the dataframe

        Args:
            nrows (int): Number of rows to print
        """
        _df = df.toPandas()
        # df_table = DFTable(_df)
        # launch df_table
        return _df


# class DFTable(QtCore.QAbstractTableModel):


#     def __init__(self, df=pd.DataFrame(), parent=None):
#         QtCore.QAbstractTableModel.__init__(self, parent=parent)
#         self._df = df.copy()

#     def toDataFrame(self):
#         return self._df.copy()

#     def headerData(self, section, orientation, role=QtCore.Qt.DisplayRole):
#         if role != QtCore.Qt.DisplayRole:
#             return QtCore.QVariant()

#         if orientation == QtCore.Qt.Horizontal:
#             try:
#                 return self._df.columns.tolist()[section]
#             except (IndexError,):
#                 return QtCore.QVariant()
#         elif orientation == QtCore.Qt.Vertical:
#             try:
#                 # return self.df.index.tolist()
#                 return self._df.index.tolist()[section]
#             except (IndexError,):
#                 return QtCore.QVariant()

#     def data(self, index, role=QtCore.Qt.DisplayRole):
#         if role != QtCore.Qt.DisplayRole:
#             return QtCore.QVariant()

#         if not index.isValid():
#             return QtCore.QVariant()

#         return QtCore.QVariant(str(self._df.ix[index.row(), index.column()]))

#     def setData(self, index, value, role):
#         row = self._df.index[index.row()]
#         col = self._df.columns[index.column()]
#         if hasattr(value, "toPyObject"):
#             # PyQt4 gets a QVariant
#             value = value.toPyObject()
#         else:
#             # PySide gets an unicode
#             dtype = self._df[col].dtype
#             if dtype != object:
#                 value = None if value == "" else dtype.type(value)
#         self._df.set_value(row, col, value)
#         return True

#     def rowCount(self, parent=QtCore.QModelIndex()):
#         return len(self._df.index)

#     def columnCount(self, parent=QtCore.QModelIndex()):
#         return len(self._df.columns)

#     def sort(self, column, order):
#         colname = self._df.columns.tolist()[column]
#         self.layoutAboutToBeChanged.emit()
#         self._df.sort_values(
#             colname, ascending=order == QtCore.Qt.AscendingOrder, inplace=True
#         )
#         self._df.reset_index(inplace=True, drop=True)
#         self.layoutChanged.emit()
